# The Gutenborg Protocol

This protocol allows you to create *GutenSites*, that is, Gopher menus that behave in a more modern manner, allowing you to define variable width fonts, headers, horizontal rules and inline images.

## The general idea

Using the Gutenborg Protocol should not make it mandatory for clients to support Gutenborg too. Current Gopher clients should be able to read GutenPages just fine, without much trouble.

Inline images, for instance, are simply `I` entries, so other clients will allow users to see the image in the way they find fit.

The rule is: **GutenPages should look nice being viewed both in Gutenborg enabled and "classic" Gopher clients**.

## Indicating Gutenborg support

Your Gopher menu must start with `+` character repeated **at least 50 times**.

The reason for that is simple: if the user client does not support Gutenborg Protocol, all he will se is a line like this on top of the current menu:

`++++++++++++++++++++++++++++++++++++++++++++++++++`

Any menu that contains such line as the first line is considered to be a GutenSite.

## Formatting

### Headers

Headers are displayed in larger fonts and their levels follow the same idea behind HTML header levels: 1 is bigger, 2 is slightly less big and so on.

```
+ Header 1

+- Header 2

+-- Header 3

+--- Header 4
```

### Horizontal rules

Any line containing `+`, `-` or `=` repeated **five or more times**, as long as the line starts with one of these characters and is composed exclusively of them, is considered an horizontal rule.

Notice that the top line that indicates the menu is a GutenSite **is** interpreted as a horizontal rule. Most of the time, you'll want to "wrap" your header between two of them, like this:

```
++++++++++++++++++++++++++++++++++++++++++++++++++
+ Title of this GutenPage
++++++++++++++++++++++++++++++++++++++++++++++++++
```

### Variable width text

Any line starting with `|` character is supposed to be displayed in a variable width font.

The idea of marking **this** case and not using, let's say, `>` to indicate fixed width fonts is that older Gopher menus can be "modernized" without too much mess: authors can enable the Gutenborg Protocol even without changing anything except the first line, and proceed to mark the lines they see fit in their own pace.

Besides, that's one reason for **not** following MarkDown syntax: if we mark "paragraphs" with `|`, the combination between paragraphs and headers become something visually more appealing for common Gopher clients:

```
+- Main section
+-- Subsection
| Here we write the text for this subsection. 
|
| See that this text **seems** to be related to the
| *Subsection* headers, as it indeed is.
```

### Fixed width fonts

Just write normal text, without beginning the lines with `|` or `+`.

As mentioned before, that allows you to migrate older Gopher menus into GutenPages in your own pace.

### Bold, italic and superscript

```
| **This** is bold.
| But *this* is italic.
| ~~And this is superscript text~~
```

#### Nested formatting

Not supported right now, but it's a planned feature.

## User interaction

We are planning a very nice format for integrating user interaction into the Gutenborg Protocol. It **probably** will use the "7 type entry" capabilities, but since the idea is **not** to break compatibility with current and older Gopher clients, we'll have to provide some way to make it happen in the easiest way possible for users of these clients.
