#!blipt
# All these lines are run inside a "namespace eval driver" command!

variable gutenborg_enabled true
variable use_gutenborg false
variable viewable_items [list]
variable download_dir "$curl::tmpdir/gopher"
variable current_state ""

# 0: Text file
proc content_handler_0 {txt domain port path} {
    variable viewable_items

    if {$path == ""} {set path "/"}

    set cmd "blipt::go_to {driver::display $domain $port {$path}}"
    blipt::fixed_link "txt→ $txt\n" $cmd
    blipt::fixed " → $domain:$port $path\n\n"

    set viewable_items [lappend viewable_items $cmd]
}
# 1: Gopher menu
proc content_handler_1 {txt domain port path} {
    global viewable_items

    if {$path == ""} {set path "/"}

    set cmd "blipt::go_to {driver::display_menu $domain $port {$path}}"
    blipt::fixed_link "$txt\n" $cmd
    blipt::fixed " → $domain:$port $path\n\n"

    set viewable_items [lappend viewable_items $cmd]
}
# 2: CCSO Nameserver
proc content_handler_2 {txt domain port path} {
    blipt::fixed "CCSO Nameserver> $txt\n"
}
# 3: Error code
proc content_handler_3 {txt domain port path} {
    blipt::fixed "ERROR> $txt\n"
}
# 4: BinHex file (download)
proc content_handler_4 {txt domain port path} {
    blipt::fixed_link "BinHex→ $txt\n" \
        "blipt::go_to {driver::display_download $domain $port {$path}}"
    blipt::fixed " → $domain:$port $path\n"
}
# 5: DOS file (download)
proc content_handler_5 {txt domain port path} {
    blipt::fixed_link "DOS→ $txt\n" \
        "blipt::go_to {driver::display_download $domain $port {$path}}"
    blipt::fixed " → $domain:$port $path\n"
}
# 6: UUEncoded file (download)
proc content_handler_6 {txt domain port path} {
    blipt::fixed_link "UUEncoded→ $txt\n" \
        "blipt::go_to {driver::display_download $domain $port {$path}}"
    blipt::fixed " → $domain:$port $path\n"
}
# 7: Search (user input)
proc content_handler_7 {txt domain port path} {
    set name "input_$blipt::elements_index"
    entry ".$name"

    blipt::fixed "$txt\n"
    $blipt::doc window create end -window ".$name"

    set cmd "driver::send_data $domain $port {$path} \[.$name get\]"
    set cmd "blipt::go_to {$cmd}"
    blipt::fixed_link " \[search\]\n" $cmd
    bind ".$name" <Return> $cmd

    blipt::fixed " → $domain:$port $path\n\n"
    blipt::newline
    incr blipt::elements_index
}
proc send_data {domain port path args} {
    set_current_place $domain $port $path

    blipt::cls; blipt::fixed "connecting to $domain:$port\n"; update
    set conn [socket $domain $port]
    chan configure $conn -buffering line

    blipt::fixed "sending request to $path\n"; update
    puts -nonewline "$path\t$args\n"
    puts -nonewline $conn "$path\t$args\n"

    blipt::title "$domain $path : $args"
    blipt::cls
    read_content $conn

    focus $blipt::doc
}

# 8: Telnet
proc content_handler_8 {txt domain port path} {
    blipt::fixed "Telnet→ $txt\n"
}
# 9: Binary file (download)
proc content_handler_9 {txt domain port path} {
    blipt::fixed_link "Bin→ $txt\n" \
        "blipt::go_to {driver::display_download $domain $port {$path}}"
    blipt::fixed " → $domain:$port $path\n"
}
# g: GIF file
proc content_handler_g {txt domain port path} {
    set current_path [get_current_path]

    set first_char [string index $path 0]
    if {$first_char == "/"} {
        set image_path "/$path"
    } else {
        set image_path "$current_path$path"
    }

    blipt::fixed_link "GIF→ $txt" \
        "blipt::go_to {driver::display_download $domain $port {$path}}"
    blipt::fixed " → $domain:$port $path\n"
}
# I: Image file
proc content_handler_I {txt domain port path} {
    variable use_gutenborg

    set first_char [string index $path 0]
    if {$first_char == "/"} {
        set image_path "/$path"
    } else {
        set current_path [get_current_path]
        set image_path "$current_path$path"
    }

    if {$use_gutenborg == false} {
        blipt::fixed_link "Image→ $txt\n" \
            "blipt::go_to {driver::display_image $domain $port {$image_path}}"
        blipt::fixed " → $domain:$port $path\n"
        return
    }

    set filepath [download $domain $port $path]
    set image_frame [blipt::img "$filepath" "$txt"]
    blipt::newline

    bind $image_frame.image <1> \
        "blipt::go_to {driver::display_image $domain $port {$image_path}}"
}
# T: Telnet 3270
proc content_handler_T {txt domain port path} {
    blipt::fixed "Telnet 3270→ $txt\n"
}
# i: Information
proc content_handler_i {txt domain port path line_counter} {
    variable gutenborg_enabled
    variable use_gutenborg

    if {$line_counter == 0} {
        set use_gutenborg [check_for_gutenborg_support $txt]
        if {$use_gutenborg == true} {
            blipt::hr
            return
        }
    }

    if {$use_gutenborg == false} {
        blipt::fixed "$txt\n"
        return
    } else {
        gutenborg_parse_line $txt
    }
}

proc check_for_gutenborg_support {txt} {
    variable gutenborg_enabled

    if {$gutenborg_enabled == true} {
        if {[regexp {\+{50,}} $txt]} {
            return true
        }
    }
    return false
}


proc gutenborg_parse_line {txt} {
    if {[regexp {(\+-*) (.+)} $txt -> level content]} {
        set counter [expr [string length $level] + 1]
        set cmd "h$counter"
        eval "blipt::$cmd {$content}"
        return
    }
    if {[regexp {(.) ?(.*)} $txt -> token content]} {
        if {$token == "|"} {
            if {$content == ""} {
                blipt::newline
            } else {
                formatted_paragraph $content
                blipt::p " "
            }
            return
        }
    }
    if {[regexp {^[\-=\+]{3,}} $txt]} {
        blipt::hr
        return
    }

    blipt::fixed "$txt\n"

}

proc formatted_paragraph {content} {
    if {[regexp {(.*)\*\*([^*]+)\*\*(.*)} $content -> p1 x p2]} {
        formatted_paragraph $p1
        blipt::bold $x
        formatted_paragraph $p2
    } elseif {[regexp {(.*)\*([^*]+)\*(.*)} $content -> p1 x p2]} {
        formatted_paragraph $p1
        blipt::italic $x
        formatted_paragraph $p2
    } elseif {[regexp {(.*)~~([^~]+)~~(.*)} $content -> p1 x p2]} {
        formatted_paragraph $p1
        blipt::overstrike $x
        formatted_paragraph $p2
    } else {
        blipt::p $content
    }
}


# s: Sound file (download)
proc content_handler_s {txt domain port path} {
    blipt::fixed_link "Sound→ $txt\n" \
        "blipt::go_to {driver::display_download $domain $port {$path}}"
    blipt::fixed " → $domain:$port $path\n"
}
# M: Mailbox???
proc content_handler_M {txt domain port path} {
    blipt::fixed_link "Mailbox→ $txt\n" \
        "blipt::go_to {driver::display_download $domain $port {$path}}"
    blipt::fixed " → $domain:$port $path\n"
}
# h: HTML!!!
proc content_handler_h {txt domain port path} {
    if {$path == ""} {set path "/"}
    blipt::fixed_link "HTML→ $txt\n" \
        "blipt::go_to {driver::display $domain $port {$path}}"
    blipt::fixed " → $domain:$port $path\n"
}
# d: Document (.doc, .pdf)
proc content_handler_d {txt domain port path} {
    blipt::fixed_link "Document→ $txt\n" \
        "blipt::go_to {driver::display $domain $port {$path}}"
    blipt::fixed " → $domain:$port $path\n"
}
# empty: something weird is happening...
proc content_handler_ {txt domain port path} {
    blipt::fixed "???→ $txt ($domain:$port/$path)"
}

proc download {domain port path} {
    variable download_dir
    # gopher://mozz.us:7003/search?tag=18&image=1/images/tennessee-rye-brooklyn.jpg
    # ↑ Wrong ↑
    # Should be: gopher://mozz.us:7003//images/tennessee-rye-brooklyn.jpg
    set parts [split $path "/"]
    set filename [lindex $parts end]

    set domain_parts [split $domain ":"]
    set l [llength $domain_parts]
    if {$l == 1} {
        set domain "gopher://$domain"
    }
    set path_and_querystring [split $path "?"]
    set path [lindex $path_and_querystring 0]
    # XXX: we should add the querystring to the end of path, right?
    set querystring [lindex $path_and_querystring 1]

    set origin "$domain:$port/$path"
    set target "$download_dir/$filename"
    return [curl::download $origin $target true]
}

#
proc read_content {conn} {
    set line_counter 0
    while {1} {
        gets $conn line
        if {$line == "." || $line == ""} {
            break
        }
        set entry_type [string index $line 0]
        set content [string range $line 1 end]
        set entries [split $content "\t"]
        lassign $entries \
            txt path domain port

        set handler_name "content_handler_$entry_type"
        set handlers_count [llength [info procs $handler_name]]
        if {$handlers_count > 0} {
            if {$entry_type == "i"} {
                $handler_name $txt $domain $port $path $line_counter
            } else {
                $handler_name $txt $domain $port $path
            }
        } else {
            puts "Unknown entry type: $entry_type"
            puts "line: $line"
        }
        incr line_counter
    }
}

proc display_content {conn} {
    variable use_gutenborg

    set content [read $conn]
    set lines [split $content "\n"]

    set first_line [lindex $lines 0]
    set use_gutenborg [check_for_gutenborg_support $first_line]

    if {$use_gutenborg == false} {
        foreach line $lines {
            blipt::fixed "$line\n"
        }
    } else {
        foreach line $lines {
            gutenborg_parse_line $line
        }
    }

    focus $blipt::doc
}

# --------------------------
# Open Gopher menus:
proc display_menu {domain port path} {
    set_current_place $domain $port $path

    global viewable_items
    set viewable_items [list]

    blipt::cls; blipt::fixed "connecting"; update
    set conn [socket $domain $port]
    chan configure $conn -buffering line

    blipt::cls; blipt::fixed "sending request"; update
    puts -nonewline $conn "$path\n"
    blipt::cls; blipt::fixed "waiting for response"; update

    blipt::title  "$domain $path"
    blipt::cls
    read_content $conn
    close $conn

    focus $blipt::doc
}
namespace export display_menu

# Display content onscreen:
proc display {domain port path} {
    set_current_place $domain $port $path

    blipt::cls; blipt::fixed "connecting"; update
    set conn [socket $domain $port]
    chan configure $conn -buffering line

    blipt::cls; blipt::fixed "sending request"; update
    puts -nonewline $conn "$path\n"
    blipt::cls; blipt::fixed "waiting for response"; update

    blipt::title "$domain $path"

    blipt::cls;
    display_content $conn
    close $conn
}
namespace export display

# Display images onscreen:
proc display_image {domain port path} {
    set_current_place $domain $port $path

    blipt::cls
    blipt::p "downloading image"
    blipt::title "$domain $path"

    set filepath [download $domain $port $path]

    blipt::cls
    blipt::h2 $domain:$port
    blipt::h1 $path
    blipt::hr
    blipt::img "$filepath"
    blipt::hr
}
namespace export display_image

proc display_download {domain port path} {
    variable download_dir

    blipt::title "$domain:$port $path"

    blipt::cls
    blipt::fixed "downloading $path\n into $download_dir\n"; update
    download $domain $port $path
    blipt::fixed "done"
}
namespace export display_image

# --------------------------
proc set_current_place {domain port path} {
    # set place [list $domain $port $path]
    blipt::set_current_place "gopher://$domain:$port$path"
}
proc get_current_place {} {
    set current_place [blipt::get_current_place]
    # gopher://domain:port/path
    #   0    1       2      3
    set parts [split $current_place "/"]
    set path [join [lrange $parts 3 end] "/"]

    set domain_and_port [lindex $parts 2]
    set parts [split $domain_and_port ":"]
    set domain [lindex $parts 0]
    set port [lindex $parts 1]
    return [list $domain $port $path]
}
proc get_current_path {} {
    set current_place [blipt::get_current_place]
    return [lindex $current_place 2]
}

# --------------------------
# Gutenborg
proc toggle_gutenborg {} {
    variable gutenborg_enabled
    if {$gutenborg_enabled == true} {
        set gutenborg_enabled false
        puts "Gutenborg disabled"
    } else {
        set gutenborg_enabled true
        puts "Gutenborg enabled"
    }
    blipt::reload
}
# --------------------------
# Keybindings:
bind $blipt::doc g {
    # Toggle Gutenborg mode
    blipt::driver::toggle_gutenborg
}
# k: go up
bind $blipt::doc k {
    set current_place [blipt::driver::get_current_place]
    set current_domain [lindex $current_place 0]
    set current_port [lindex $current_place 1]
    set current_path [lindex $current_place 2]

    set parts [split [string trimright $current_path "/"] "/"]
    set new_path [join [lrange $parts 0 end-1] "/"]
    if {$new_path == ""} {set new_path "/"}
    blipt::go_to "blipt::driver::display_menu $current_domain $current_port {$new_path}"
}
# p: print details about current page
bind $blipt::doc p {
    set current_place [blipt::driver::get_current_place]
    set current_domain [lindex $current_place 0]
    set current_port [lindex $current_place 1]
    set current_path [lindex $current_place 2]

    set address "gopher://$current_domain:$current_port/$current_path"
    puts $address

    clipboard clear
    clipboard append -format STRING $address
}
bind $blipt::doc "\[" {
    # Go to previous viewable item
    set current_cmd [lindex $blipt::history end]
    set current_cmd "blipt::go_to {$current_cmd}"

    set idx 0
    foreach item $blipt::driver::viewable_items {
        if {$item == $current_cmd} {
            if {$idx == 0} {return}
            set previous_cmd [lindex $blipt::driver::viewable_items $idx-1]
            eval $previous_cmd
            return
        }
        incr idx
    }
}
bind $blipt::doc "\]" {
    # Go to next viewable item
    set current_cmd [lindex $blipt::history end]
    set current_cmd "blipt::go_to {$current_cmd}"

    set idx 0
    foreach item $blipt::driver::viewable_items {
        if {$item == $current_cmd} {
            set next_cmd [lindex $blipt::driver::viewable_items $idx+1]
            if {$next_cmd != ""} {
                eval $next_cmd
            }
            return
        }
        incr idx
    }
}

# --------------------------
# Main:
proc init {argv} {
    variable download_dir
    file mkdir $download_dir

    set target [lindex $argv 0]

    # Must get rid of protocol part:
    # Example: "gopher://zaibatsu.circumlunar.space:70/~sloum/"
    set idx [string first "://" $target]
    if {$idx != -1} {
        set protocol [string range $target 0 [expr $idx - 1]]
        if {$protocol != "gopher"} {
            puts "Invalid protocol: '$protocol'."
            puts "Should be 'gopher://'."
            exit 1
        }
        set target [string range $target [expr $idx + 3] end]
    }

    set parts [split $target "/"]

    set domain_and_port [split [lindex $parts 0] ":"]

    set domain [lindex $domain_and_port 0]
    if {$domain == ""} {set domain sdf.org}

    set port [lindex $domain_and_port 1]
    if {$port == ""} {set port 70}

    set path [join [lrange $parts 1 end] "/"]
    if {$path == ""} {set path "/"}

    set cmd "driver::display_menu $domain $port {$path}"
    blipt::go_to $cmd
}
